from functools import cached_property
from typing import List, Union

from django.db.models import QuerySet, OrderBy, F

from ariadne_django_resolvers.settings import api_settings
from ariadne_django_resolvers.utils import apply_cursor_to_queryset
from .edges import Edge
from .page_info import PageInfo


class Connection:
    def __init__(
        self,
        qs: QuerySet,
        first: int = None,
        last: int = None,
        before: int = None,
        after: int = None,
    ):
        self._qs = self._order_queryset(qs)
        self._first = first
        self._last = last
        self._before = before
        self._after = after

    @property
    def page_info(self):
        return PageInfo(
            qs=self._qs, edges=self.edges, before=self._before, after=self._after
        )

    @property
    def total_count(self):
        return self._qs.count()

    # Cached because the edges are used inside the page_info as well.
    @cached_property
    def edges(self) -> List[Edge]:
        qs = self._qs

        if self._before:
            qs = apply_cursor_to_queryset(qs=qs, before=self._before)
            qs = qs.exclude(pk=self._before)

        if self._after:
            qs = apply_cursor_to_queryset(qs=qs, after=self._after)
            qs = qs.exclude(pk=self._after)

        if (
            not self._last
            and not self._first
            and qs.count() > api_settings.MAX_PAGE_SIZE
        ):
            raise ValueError("You need to either specify first or last.")
        elif self._last and self._first:
            raise ValueError("Do not specify first and last together.")
        elif any(
            (
                self._first and self._first > api_settings.MAX_PAGE_SIZE,
                self._last and self._last > api_settings.MAX_PAGE_SIZE,
            )
        ):
            raise ValueError(f"The maximum page size is {api_settings.MAX_PAGE_SIZE}.")
        elif self._first is not None:
            qs = qs[: self._first]
        elif self._last is not None:
            # Negative indexing is not supported, so reverse the queryset and get last
            # <last> objects.
            # Then reverse the queryset again, to retain the order.
            # Calling .reverse again does not work after a slice has been taken,
            # so use reversed instead.
            # After this the obejct is no QuerySet anymore, but it does not matter,
            # because then at this point you only need to be able to iterate through it.
            qs = reversed(qs.reverse()[: self._last])

        return [Edge(obj=obj) for obj in qs]

    @staticmethod
    def _order_queryset(qs):
        # Make sure the queryset is always ordered by the primary key (at least as the
        # last specified order field)
        # to make sure the pagination works at all times.
        # Because:
        #   1. If the queryset was not ordered at all, the returned edges would be
        #       totally random.
        #   2. If the queryset was ordered, but the value of an object was the same
        #       as the value of another object the returned edges would be again a
        #       little random.
        ordering = []

        def format_ordering(order_by: Union[str, OrderBy]):
            if isinstance(order_by, OrderBy):
                if order_by.nulls_first:
                    # Prevent 'nulls_first' to have a consistent way to paginate.
                    raise ValueError(
                        "Ordered QuerySets with 'nulls_first' set to True are "
                        "not supported."
                    )
                order_by.nulls_last = True
                return order_by
            else:
                if order_by.startswith("-"):
                    return F(order_by[1:]).desc(nulls_last=True)
                else:
                    return F(order_by).asc(nulls_last=True)

        if qs.query.extra_order_by:
            ordering = [
                format_ordering(order_by=order_by)
                for order_by in qs.query.extra_order_by
            ]
        elif qs.query.order_by:
            ordering = [
                format_ordering(order_by=order_by) for order_by in qs.query.order_by
            ]
        elif qs.query.get_meta().ordering:
            ordering = [
                format_ordering(order_by=order_by)
                for order_by in qs.query.get_meta().ordering
            ]

        return qs.order_by(*ordering, "pk")
