from typing import Any

from django.db.models import QuerySet, Model
from graphql import GraphQLResolveInfo


class Resolver:
    queryset: QuerySet = None

    # If you want to use object lookups other than pk, set 'lookup_field'.
    # For more complex lookup requirements override `get_object()`.
    lookup_field = "pk"
    lookup_variable_name = None

    def __call__(self, parent: Any, info: GraphQLResolveInfo, **variables) -> Any:
        self.parent = parent
        self.info = info
        self.request = info.context["request"]
        self.user = self.request.user
        self.variables = variables
        return self.resolve()

    def resolve(self) -> Any:
        raise NotImplementedError()

    def get_queryset(self) -> QuerySet:
        """
        Get the list of items for this resolver.
        Defaults to using `self.queryset`.

        This method should always be used rather than accessing `self.queryset`
        directly, as `self.queryset` gets evaluated only once, and those results
        are cached for all subsequent requests.

        You may want to override this if you need to provide different
        querysets depending on the incoming request.

        (Eg. return a list of items that is specific to the user)
        """
        assert self.queryset is not None, (
            f"'{self.__class__.__name__}' should either include a `queryset` "
            "attribute, or override the `get_queryset()` method."
        )
        queryset = self.queryset

        if isinstance(queryset, QuerySet):
            # Ensure queryset is re-evaluated on each request.
            queryset = queryset.all()

        return queryset

    def get_object(self) -> Model:
        """
        Returns the object the resolver is displaying.

        You may want to override this if you need to provide non-standard
        queryset lookups.  Eg if objects are referenced using multiple
        keyword arguments.
        """
        queryset = self.get_queryset()

        lookup_variable_name = self.lookup_variable_name or self.lookup_field

        assert lookup_variable_name in self.variables, (
            f"Expected resolver {self.__class__.__name__} to be called with a variable "
            f"named '{lookup_variable_name}. "
            "Fix your Query field, or set the `.lookup_field` attribute on the resolver"
            " correctly."
        )

        filter_kwargs = {self.lookup_field: self.variables[lookup_variable_name]}

        obj = queryset.get(**filter_kwargs)

        return obj
