from typing import List

from django.db.models import QuerySet, F, OrderBy

from .base import Resolver
from ..utils import convert_camel_case_to_snake_case


class ListResolver(Resolver):
    filters_variable_name = "filters"
    order_by_variable_name = "order_by"

    def resolve(self) -> QuerySet:
        queryset = self.get_queryset()
        queryset = self.filter_queryset(queryset)
        queryset = self.order_queryset(queryset)
        return queryset

    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        filters = self.variables.get(self.filters_variable_name, {})
        return queryset.filter(**filters)

    def order_queryset(self, queryset: QuerySet) -> QuerySet:
        order_by = self.variables.get(self.order_by_variable_name, [])
        order_by = self._format_order_by(order_by)
        return queryset.order_by(*order_by)

    @staticmethod
    def _format_order_by(order_by: List[dict]) -> List[OrderBy]:
        """
        Convert your ordering dictionaries to 'OrderBy' objects:

        Info:
            None values are always sorted last for consistency.

        Args:
            order_by: A list of dictionaries with the keys 'field' (str) and
                'desc' (bool).

        Returns: A list of 'OrderBy' objects' as its expected by the QuerySet's
            '.order_by' method.

        Examples:
            Input:
                [{
                    "field": "name",
                    "desc": True
                }]
            Output:
                [F("name").desc(nulls_last=True)]
        """

        def format_order_field_name(order_dict: dict) -> OrderBy:
            field_name = convert_camel_case_to_snake_case(order_dict["field"])
            if order_dict["desc"]:
                return F(field_name).desc(nulls_last=True)
            else:
                return F(field_name).asc(nulls_last=True)

        return [format_order_field_name(order_dict) for order_dict in order_by]
