from typing import Optional, List

from django.contrib.auth import get_user_model
from django.db.models import QuerySet
from django.urls import reverse

from ariadne_django_resolvers.settings import api_settings

User = get_user_model()

UNSET = "UNSET"
id_scalar = api_settings.SCHEMA.get_type("ID")


# noinspection PyUnresolvedReferences
class ResolverTestMixin:
    query: str

    login_required: bool = False

    mutation_name: str
    response_typename: str = "MutationResponse"

    default_variables: dict = {}

    def run_query(self, user_id: Optional[int] = UNSET, **variables):
        if user_id == UNSET and self.login_required:
            user_id = 1

        if user_id and user_id != UNSET:
            user = User.objects.get(pk=user_id)
            self.client.force_login(user=user)

        return self.client.post(
            path=reverse("graphql"),
            data={
                "query": self.query,
                "variables": variables or self.default_variables,
            },
            content_type="application/json",
        )

    def assertObjectDoesNotExistErrorRaised(self, response):
        errors = response.json()["errors"]
        self.assertTrue(
            any(
                (
                    "matching query does not exist." in error["message"]
                    for error in errors
                )
            )
        )

    def assertMutationResponseValid(self, response, ok: bool):
        """Test the __typename, ok and the errors field."""
        data = response.json()["data"][self.mutation_name]
        self.assertEqual(self.response_typename, data["__typename"])

        if ok:
            self.assertTrue(data["ok"])
            self.assertEqual([], data["errors"])
        else:
            self.assertFalse(data["ok"])
            self.assertTrue(len(data["errors"]) > 0)

    def assertObjectDoesNotExist(self, queryset: QuerySet, pk: int):
        self.assertFalse(queryset.filter(pk=pk).exists())

    @staticmethod
    def get_edge_by_node_id(connection: dict, node_id: str):
        return next(
            (edge for edge in connection["edges"] if edge["node"]["id"] == node_id)
        )

    @staticmethod
    def get_node_ids_from_connection(connection: dict) -> List[int]:
        return [
            id_scalar.parse_value(edge["node"]["id"]) for edge in connection["edges"]
        ]
