import asyncio
from functools import wraps
from typing import List, Any, Tuple, Iterator, Callable, Dict

from django.db.models import QuerySet, Model, OrderBy, Q


def convert_kwargs_to_snake_case(func: Callable) -> Callable:  # pragma: no cover
    # Wait for ariadne 0.13 to be released, then you should be able to remove this.
    def convert_to_snake_case(d: Dict) -> Dict:
        converted: Dict = {}
        for k, v in d.items():
            if isinstance(v, dict):
                v = convert_to_snake_case(v)
            if isinstance(v, list):
                v = [convert_to_snake_case(i) if isinstance(i, dict) else i for i in v]
            converted[convert_camel_case_to_snake_case(k)] = v
        return converted

    if asyncio.iscoroutinefunction(func):

        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **convert_to_snake_case(kwargs))

        return async_wrapper

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return func(*args, **convert_to_snake_case(kwargs))

    return wrapper


def convert_camel_case_to_snake_case(camel_case_string: str) -> str:  # pragma: no cover
    """
    Convert a string in camelCase to snake_case.
    Args:
        camel_case_string: A string in camelCase.

    Returns: The given <camel_case_string> converted to snake_case.

    """
    snake_case_string = ""

    for i, char in enumerate(camel_case_string.lower()):
        if char != camel_case_string[i] and i > 0 and camel_case_string[i - 1] != "_":
            snake_case_string += "_"
        snake_case_string += char

    return snake_case_string


def convert_snake_case_to_camel_case(snake_case_string: str) -> str:
    """
    Convert a string from snake_case to camelCase.
    Only single underscores cause a removal of it and an conversion of the following
    letter to uppercase, multiple underscores in a row will be preserved.

    Examples:
        field_id => fieldId
        field__id => field__id

    Args:
        snake_case_string: A string in snake_case.

    Returns: The given <snake_case_string> converted to camelCase.

    """
    camel_string = ""

    for i, c in enumerate(snake_case_string):
        if (
            c == "_"
            and i != 0
            and i != len(snake_case_string) - 1
            and snake_case_string[i - 1] != "_"
            and snake_case_string[i + 1] != "_"
        ):
            continue
        elif (
            i > 0
            and snake_case_string[i - 1] == "_"
            and snake_case_string[i - 2] != "_"
            and i != 1
        ):
            c = c.upper()
        camel_string += c
    return camel_string


def get_nested_object_attribute(obj: Model, field: str):
    """
    Get the value of a nested field from an object.

    Args:
        obj: An instance of a django model.
        field: A string like 'restaurant__adress__zip_code' or just 'name'.

    Returns:
        The value of that nested attribute.

    """
    if "__" in field:
        field, field_rest = field.split("__", 1)
    else:
        field_rest = ""

    if field_rest:
        return get_nested_object_attribute(getattr(obj, field), field_rest)
    else:
        return getattr(obj, field)


def get_page_filters_from_order_by(
    reference_object: Model, order_by: List[str], reverse_lookups: bool
) -> Iterator[Tuple[str, str, Any]]:
    """
    Convert an ordering to the corresponding filters you'd need to get all object after
    of before the reference_object in a queryset.

    Args:
        reference_object: The reference object (eg. the object retrieved from a given
            cursor).
        order_by: The queryset's current ordering.
        reverse_lookups: A boolean indicating whether the filters should be reversed.
            (True if used for before)

    Returns: The filters to retrieve the objects after or before the given
        reference_object.

    """
    for ordering in order_by:
        if isinstance(ordering, OrderBy):
            field = ordering.expression.name
            if ordering.descending:
                lookup_expr = "gt"
            else:
                lookup_expr = "lt"
        else:
            field = ordering
            if field.startswith("-"):
                field = field[1:]
                lookup_expr = "gt"
            else:
                lookup_expr = "lt"

        if reverse_lookups:
            lookup_expr = "gt" if lookup_expr == "lt" else "lt"

        value = get_nested_object_attribute(reference_object, field)

        yield field, lookup_expr, value


def apply_cursor_to_queryset(
    qs: QuerySet, after: int = None, before: int = None,
) -> QuerySet:
    """
    Return the objects coming in the queryset after or before the given primary key
        (after or before).

    Args:
        qs: The queryset to filter.
        after: An object id whose values will be used for filtering.
        before: An object id whose values will be used for filtering.

    Returns: The filtered queryset.

    """
    if not after and not before:
        raise ValueError(
            "You need to specify either 'after' or the 'before' parameter."
        )
    elif after and before:
        raise ValueError("You can only specify either 'after' or 'before'.")
    elif qs.ordered is False:
        raise ValueError("The QuerySet must be ordered.")

    pk = after if after else before
    reference_object = qs.get(pk=pk)
    order_by = qs.query.order_by
    filters = get_page_filters_from_order_by(
        reference_object=reference_object, order_by=order_by, reverse_lookups=not before
    )

    for (field, lookup, value) in filters:
        # Unless the primary key is not the only field the queryset is ordered by,
        # do not use it to
        # filter, because it's function as a secondary filter is already handled.
        # And forcing a pk filter would return a wrong queryset.
        # PS: Because you're forcing all querysets to be ordered by the pk
        # (in the Connection's '_order_queryset' method) it's the only field that could
        # be set standalone so it's valid  to just use len(order_by) != 1.
        if field == "pk" and len(order_by) != 1:
            continue

        if value is None:
            f = Q(**{f"{field}__isnull": True}) & Q(
                **{f"pk__{'gt' if after else 'lt'}": pk}
            )
            if before:
                f |= Q(**{f"{field}__isnull": False})

            qs = qs.filter(f)
        else:
            # Filter either by 'gt' or 'lt' than value or for objects with the same
            # value but with a higher or lower primary key
            # (depending on the reverse variable).
            f = Q(**{f"{field}__{lookup}": value})
            if field != "pk":
                f |= Q(**{field: value}) & Q(**{f"pk__{'gt' if after else 'lt'}": pk})
            if after:
                f |= Q(**{f"{field}__isnull": True})

            qs = qs.filter(f)

    return qs
